/*
 *  File:             tal.c
 *  Date created:     April 25, 1997 (Friday, 20:12h)
 *  Author:           Thomas Jensen
 *                    tsjensen@cip.informatik.uni-erlangen.de
 *  Version:          $Id: tal.c,v 1.5 1997/06/30 02:42:56 l95coop4 Exp $
 *  Language:         ANSI C
 *  Platforms tested: AIX 4.1, SUNOS 5, IRIX 5, IRIX 6, HPUX 9, HPUX10, alpha,
 *                    and Linux. Doesn't seem to work sunos4 (no getopt()).
 *  World Wide Web:   http://home.pages.de/~jensen/tal/
 *  Purpose:          Align common characters at the end of lines read from
 *                    stdin. Output goes to stdout.
 *  Remarks:          The program could certainly be sped up. I didn't put
 *                    much effort into improving performance, because in
 *                    almost all cases, the number of processed lines is
 *                    less then one 'screenful'.
 *  Revision History:
 *
 *    $Log: tal.c,v $
 * Revision 1.5  1997/06/30  02:42:56  l95coop4
 * 1. Changed padding default value to 1
 * 2. Taken out trailer frequency/tolerance options
 * 3. Fixed bug: Padding was added even if trailer frequency was 1
 * 4. Added bounds check to trailing whitespace removal
 * 5. Increased maximum allowed tab stop distance to 40
 * 6. If it doesn't conflict with -l, we now choose a shorter trailer
 *    if the current one starts with a fillchar (improves \ handling)
 *
 *    Revision 1.4  1997/05/09  01:06:39  l95coop4
 *    Port to sunos5, irix5, irix6, hpux9, hpux10, and alpha
 *
 *    Revision 1.3  1997/05/03  02:27:24  l95coop4
 *    Changed some of the default values
 *    Added padding option, added fixed column option
 *    Added print version option, added force option
 *
 *    Revision 1.2  1997/05/02  03:13:15  l95coop4
 *    Added tab handling. Now, only tabs that are after the end of the text
 *    (but maybe before the trailer) are expanded. All others are preserved.
 *    Fixed bug connected to the specification of invalid options.
 *
 *    Revision 1.1  1997/04/27  20:05:16  l95coop4
 *    Initial revision
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>


const char rcsid[] =
   "$Id: tal.c,v 1.5 1997/06/30 02:42:56 l95coop4 Exp $";


#define PROJECT            "tal"         /* name of the program */

#define MAX_TRAILERLEN     200           /* maximum supported trailer length */
#define INPUTBUF_LEN       2048          /* maximum supported line length */

#define DEF_FILLCHAR       ' '           /* default fill character */
#define DEF_PREFTLEN       0             /* default preferred trailer length */
#define DEF_MAXALIGN       87            /* default maximum desired line length */
#define DEF_TABSTOP        8             /* default tab stop distance */
#define DEF_COLUMN         0             /* default end column */
#define DEF_PADDING        1             /* default number of chars bef. trailer */
#define DEF_FORCE_TLEN     false         /* default strictness of pref_tlen */


#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1                   /* Tsk. Not defined on some systems. */
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif

typedef enum { false, true } bool;       /* Boolean data type */


typedef struct {                         /* Line to be processed: */
   char *text;                           /* contents of line (tabs expanded) */
   int   len;                            /* length of the contents (w/o \n) */
   char *tend;                           /* pointer to last char of orgtext */
   bool  active;                         /* true if line is considered */
   char *orgtext;                        /* contents of line (tabs not exp.) */
} line_t;

line_t **lines;                          /* the actual input */
unsigned long lines_size;                /* number of line pointers reserved */

unsigned long anz_lines;                 /* number of lines read */


typedef struct {                         /* Trailer from list */
   char *text;                           /* pointer to trailer text */
   int   len;                            /* length of that trailer */
   unsigned long freq;                   /* absolute frequency */
} trailer_t;

trailer_t **trailers;                    /* list of trailer candidates */
unsigned long trailers_size;             /* num of trailer pointers reserved */

unsigned long anz_trailers;              /* how much of the list is in use */

trailer_t *used_trail = NULL;            /* The trailer that's actually used. */

bool do_nothing = false;                 /* true, if we should do nothing */

struct {                                 /* Command Line Options: */
   int  column;                          /* Trailer must end on that column */
   char fillchar;                        /* used for filling up the new space */
   bool force_tlen;                      /* choose trailer of length pref_tlen */
   int  maxalign;                        /* maximum desired line length */
   int  padding;                         /* number of fillchars inserted before trailer */
   int  pref_tlen;                       /* preferred trailer length (0 == disable) */
   int  tabstop;                         /* distance between tab stops */
} opt;




/****************************************************************************/

void usage()
/*
 *  Display usage information.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   printf ("Usage: %s [options] < input > output\n", PROJECT);
   printf ("       -c char  fill character [def.: '%c']\n", DEF_FILLCHAR);
   printf ("       -f       force preferred trailer length [def.: %s]\n", DEF_FORCE_TLEN? "on":"off");
   printf ("       -h       usage info\n");
   printf ("       -k uint  align trailers to end at this column (0 == disable) [def.: %d]\n", DEF_COLUMN);
   printf ("       -l uint  preferred trailer length (0 == disable) [def.: %d]\n", DEF_PREFTLEN);
   printf ("       -m uint  maximum length you want a line to reach [def.: %d]\n", DEF_MAXALIGN);
   printf ("       -p uint  number of chars inserted before trailer [def.: %d]\n", DEF_PADDING);
   printf ("       -t uint  distance between tab stops [def.: %d]\n", DEF_TABSTOP);
   printf ("       -v       print out current version number\n");
}    



/****************************************************************************/

void process_commandline (int argc, char *argv[])
/*
 *  Process command line options and set global options in opt.
 *  If necessary, usage() is displayed.
 *
 *  argc, argv     Command line as received by main()
 *
 *  When this function returns, the command line has been correct
 *  and opt has been set. Otherwise, the entire program is terminated.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   int oc;                               /* option character */
   int idummy;

   opt.column     = DEF_COLUMN;          /* set the defaults */
   opt.fillchar   = DEF_FILLCHAR;
   opt.force_tlen = DEF_FORCE_TLEN;
   opt.maxalign   = DEF_MAXALIGN;
   opt.padding    = DEF_PADDING;
   opt.pref_tlen  = DEF_PREFTLEN;
   opt.tabstop    = DEF_TABSTOP;

   do {
      oc = getopt (argc, argv, "c:fhk:l:m:p:t:v");

      switch (oc) {

         case 'c':
            /*
             *  Fill character
             */
            if (strlen (optarg) != 1) {
               fprintf (stderr,
                  "%s: Fill char must be exactly one character.\n", PROJECT);
               exit (EXIT_FAILURE);
            }
            opt.fillchar = optarg[0];
            break;

         case 'l':
            /*
             *  Preferred trailer length
             */
            idummy = (int) strtol (optarg, NULL, 10);
            if (idummy < 0 || idummy > MAX_TRAILERLEN) {
               fprintf (stderr,
                  "%s: Preferred trailer length must be between 0 and %d.\n",
                  PROJECT, MAX_TRAILERLEN);
               exit (EXIT_FAILURE);
            }
            opt.pref_tlen = idummy;
            break;

         case 'm':
            /*
             *  Maximum desired line length
             */
            idummy = (int) strtol (optarg, NULL, 10);
            if (idummy < 0) {
               fprintf (stderr,
                  "%s: Negative line lengths not allowed.\n", PROJECT);
               exit (EXIT_FAILURE);
            }
            opt.maxalign = idummy;
            break;

         case 't':
            /*
             *  Distance between tab stops
             */
            idummy = (int) strtol (optarg, NULL, 10);
            if (idummy < 1 || idummy > 40) {
               fprintf (stderr,
                  "%s: '%d' is an invalid tab stop distance.\n",
                  PROJECT, idummy);
               exit (EXIT_FAILURE);
            }
            opt.tabstop = idummy;
            break;

         case 'k':
            /*
             *  Make all trailers end at a particular column
             */
            idummy = (int) strtol (optarg, NULL, 10);
            if (idummy < 0 || idummy > INPUTBUF_LEN) {
               fprintf (stderr,
                  "%s: '%d' is an invalid trailer ending column.\n",
                  PROJECT, idummy);
               exit (EXIT_FAILURE);
            }
            opt.column = idummy;
            break;

         case 'p':
            /*
             *  Padding
             */
            idummy = (int) strtol (optarg, NULL, 10);
            if (idummy < 0 || idummy > INPUTBUF_LEN) {
               fprintf (stderr,
                  "%s: '%d' is an invalid padding value.\n",
                  PROJECT, idummy);
               exit (EXIT_FAILURE);
            }
            opt.padding = idummy;
            break;

         case 'f':
            /*
             *  Force preferred trailer length
             */
            opt.force_tlen = true;
            break;

         case 'v':
            /*
             *  Display current version
             */
            printf ("%s - trailer alignment ", PROJECT);
            printf ("(aligns common characters at the end of lines)\n");
            printf ("%s\n", rcsid);
            exit (EXIT_SUCCESS);

         case 'h':
            /*
             *  Display usage info
             */
            printf ("%s -  trailer alignment ", PROJECT);
            printf ("(aligns common characters at the end of lines)\n");
            printf ("       (c) 1997 Thomas Jensen ");
            printf ("<tsjensen@cip.informatik.uni-erlangen.de>\n");
            usage();
            exit (EXIT_SUCCESS);

         case ':': case '?':
            /*
             *  Missing argument or illegal option
             */
            usage();
            exit (EXIT_SUCCESS);

         case EOF:
            /*
             *  End of list, do nothing more
             */
            break;

         default:                        /* This case must never be */
             fprintf (stderr,
                "%s: I'm sorry to say that my author was an incompetent jerk.\n",
                PROJECT);
             exit (EXIT_FAILURE);
      }

   } while (oc != EOF);
}


/****************************************************************************/

int expand_tabs_into (const char *input_buffer, char **text)
/*
 *  Expand tab chars in input_buffer and store result in text.
 *
 *  input_buffer   Line of text with tab chars
 *  text           address of the pointer that will take the result
 *
 *  Memory will be allocated for the result. Returns only if successful.
 *
 *  Returns:  Length of the result line in characters
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   char *temp;                           /* work string (more memory) */
   int in_len = strlen (input_buffer);   /* length of input string */
   int ii;                               /* position in input string */
   int io;                               /* position in work string */
   int jp;                               /* tab expansion jump point */

   temp = (char *) malloc (INPUTBUF_LEN * opt.tabstop);
   if (temp == NULL) {
      perror (PROJECT);
      exit (EXIT_FAILURE);
   }
   for (ii=0, io=0; ii<in_len && io<(INPUTBUF_LEN*opt.tabstop-1); ++ii) {
      if (input_buffer[ii] == '\t') {
         for (jp=io+opt.tabstop-(io%opt.tabstop); io<jp; ++io)
            temp[io] = ' ';
      }
      else {
         temp[io] = input_buffer[ii];
         ++io;
      }
   }
   temp[io] = '\0';

   *text = (char *) strdup (temp);
   if (*text == NULL) {
      perror (PROJECT);
      exit (EXIT_FAILURE);
   }
   free (temp);

   return strlen (*text);
}


/****************************************************************************/

void read_all_input()
/*
 *  Read all input lines from stdin and build lines data structures.
 *
 *  When this function returns, the lines data structures have been
 *  successfully built. Otherwise, the entire program is terminated.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   char input_buffer[INPUTBUF_LEN];      /* where we store each line for processing */
   char *p;
   int in_len;                           /* length of string in input_buffer */

   anz_lines = 0;                        /* no lines read yet */
   lines = (line_t **) calloc (100, sizeof (line_t *));
   if (lines == NULL) {
      perror (PROJECT);
      exit (EXIT_FAILURE);
   }
   lines_size = 100;                     /* allocated 100 line pointers */

   while (fgets (input_buffer, INPUTBUF_LEN, stdin)) {

      /*
       *  If needed, expand line pointer array
       */
      if (anz_lines == lines_size) {
         lines_size += 100;
         lines = (line_t **) realloc (lines, lines_size * sizeof (line_t *));
         if (lines == NULL) {
            perror (PROJECT);
            exit (EXIT_FAILURE);
         }
      }
      /*
       *  Allocate line structure and check if input present
       */
      lines[anz_lines] = (line_t *) calloc (1, sizeof (line_t));
      if (lines[anz_lines] == NULL) {
         perror (PROJECT);
         exit (EXIT_FAILURE);
      }
      in_len = strlen (input_buffer);
      if (in_len == 0) {
         /* There must be some read error, because we should
          * at least have received a newline. */
         free (lines[anz_lines]);
         lines[anz_lines] = NULL;
         break;
      }
      /*
       *  Remove trailing whitespace from input buffer, then store length
       */
      p = input_buffer + in_len - 1;
      while (p >= input_buffer && (*p == ' ' || *p == '\t' || *p == '\n'))
         *p-- = '\0';
      lines[anz_lines]->len = strlen (input_buffer);
      /*
       *  Store line in lines list
       */
      lines[anz_lines]->orgtext = (char *) strdup (input_buffer);
      lines[anz_lines]->text = NULL;
      lines[anz_lines]->tend = (lines[anz_lines]->orgtext)+(lines[anz_lines]->len)-1;
      /*
       *  Was it an empty line?
       */
      if (lines[anz_lines]->orgtext[0] == '\0') {
         lines[anz_lines]->tend = lines[anz_lines]->orgtext;
         lines[anz_lines]->len = 0;
         lines[anz_lines]->active = false;
         ++anz_lines;
         continue;
      }
      /*
       *  A normal line. Next one please.
       */
      lines[anz_lines]->active = true;
      ++anz_lines;
   }
}


/****************************************************************************/

int longest_line_len()
/*
 *  Returns length of longest line in characters.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   unsigned long i;
   int max = 0;

   for (i=0; i<anz_lines; ++i) {
      if (lines[i]->active && lines[i]->len > max)
         max = lines[i]->len;
   }
   return max;
}


/****************************************************************************/

trailer_t *texist (trailer_t **temple,
                   const char *wanted,
                   const unsigned long anz_temple)
/*
 *  Look for the wanted trailer in the temporary list of trailers.
 *
 *  temple        temporary list of trailers (all same length)
 *  wanted        trailer we want to locate
 *  anz_temple    number of entries in temporary list
 *
 *  Returns:  == NULL   trailer not found
 *            != NULL   pointer to existing trailer
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   unsigned long i;

   for (i=0; i<anz_temple; ++i) {
      if (strcmp (temple[i]->text, wanted) == 0)
         return (trailer_t *) (temple[i]);
   }
   return NULL;
}


/****************************************************************************/

void tadd (trailer_t         **temple,
           const unsigned long anz_temple,
           const char         *tstart)
/*
 *  Add the trailer tstart to the temporary list.
 *
 *  temple        temporary list of trailers (all same length)
 *  tstart        trailer we want to add to the list
 *  anz_temple    number of entries in temporary list
 *
 *  Returns only when successful. No adjustments of anz_temple etc. is done.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   temple[anz_temple] = (trailer_t *) calloc (1, sizeof(trailer_t));
   if (temple[anz_temple] == NULL) {
      perror (PROJECT);
      exit (EXIT_FAILURE);
   }
   temple[anz_temple]->text = (char *) tstart;
   temple[anz_temple]->len = strlen (tstart);
   temple[anz_temple]->freq = 1;
}


/****************************************************************************/

unsigned long count_active_lines()
/*
 *  Count lines in lines list that are tagged active.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   unsigned long count = 0;
   unsigned long i;

   for (i=0; i<anz_lines; ++i)
      if (lines[i]->active) ++count;
   return count;
}


/****************************************************************************/

void find_a_good_trailer()
/*
 *  This will examine the input lines and try to find a good trailer.
 *
 *  When this function returns, the trailer to be used will be stored in
 *  used_trail. If we absolutely couldn't find a suitable trailer, the
 *  entire program is terminated.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   int longest = longest_line_len();     /* length of longest line */
   int cur_tlen;                         /* currently considered trailer len */
   trailer_t **temple;                   /* start of temporary list entries */
   unsigned long temple_space;           /* space left for temp list */
   unsigned long anz_temple;             /* number of entries in temple */
   line_t *p;                            /* roving line pointer */
   unsigned long i;                      /* loop counter */
   trailer_t *tp;
   unsigned long tmostfreq;              /* temple index to most frequent temp trailer */
   unsigned long anz_valines =
      count_active_lines();              /* number of considered lines */

   if (anz_valines == 0) return;
   /*
    *  Start by allocating enough memory for the completed trailer list
    */
   trailers = (trailer_t **) calloc (longest, sizeof (trailer_t *));
   if (trailers == NULL) {
      perror (PROJECT);
      exit (EXIT_FAILURE);
   }
   trailers_size = longest;
   anz_trailers = 0;

   for (cur_tlen=1; cur_tlen<=longest; ++cur_tlen) {
      
      temple = trailers + (cur_tlen-1);
      temple_space = trailers_size - (cur_tlen-1);
      anz_temple = 0;

      /*
       *  Create list of all possible trailers of the current length
       */
      for (i=0, p=lines[0]; i<anz_lines; p=lines[++i]) {

         if (p->active == false) continue;
         if (p->len < cur_tlen) continue;

         tp = texist (temple, (p->tend)-cur_tlen+1, anz_temple);
         if (tp != NULL) {
            ++(tp->freq);
            continue;
         }

         if (temple_space <= 0) {
            /*
             *  Increase trailer list size
             */
            trailers_size += 100;
            trailers = (trailer_t **) realloc (trailers, trailers_size * sizeof (trailer_t *));
            if (trailers == NULL) {
               perror (PROJECT);
               exit (EXIT_FAILURE);
            }
            temple_space += 100;
         }
         tadd (temple, anz_temple, (p->tend)-cur_tlen+1);
         --temple_space;
         ++anz_temple;
      }

      /*
       *  From the temporary list, choose the most frequent trailer
       *  and put it up front. Then delete all other candidates.
       */
      tmostfreq = 0;
      for (i=1; i<anz_temple; ++i) {
         if (temple[i]->freq > temple[tmostfreq]->freq)
            tmostfreq = i;
      }
      tp = temple[0];
      temple[0] = temple[tmostfreq];
      temple[tmostfreq] = tp;
      for (i=1; i<anz_temple; ++i) {
         free (temple[i]);
         temple[i] = NULL;
      }
      ++anz_trailers;
   }

   /*
    *  At this point, the trailer list contains for every possible trailer
    *  length the most frequent trailer of that length and its text, plus
    *  information about its actual frequency value.
    */

   /* for (i=0; i<longest; ++i) {                                       */
   /*    printf ("%2d: len:%d, freq:%ld, \"%s\"\n",                     */
   /*       i, trailers[i]->len, trailers[i]->freq, trailers[i]->text); */
   /* }                                                                 */
   /* exit (EXIT_SUCCESS);                                              */

   used_trail = (trailer_t *) calloc (1, sizeof (trailer_t));
   if (used_trail == NULL) {
      perror (PROJECT);
      exit (EXIT_FAILURE);
   }
   /*
    *  Choose the most frequent trailer
    */
   tmostfreq = 0;
   for (cur_tlen=1; cur_tlen<longest; ++cur_tlen) {
      if (trailers[cur_tlen]->freq >= trailers[tmostfreq]->freq)
         tmostfreq = cur_tlen;
   }
   /*
    *  If the chosen trailer is longer than preferred, take the preferred size
    */
   if (opt.pref_tlen != 0) {               /* 0 means disable preference */
      if (opt.force_tlen) {                /* we *must* take the preferred one */
         if (opt.pref_tlen > anz_trailers)
            tmostfreq = anz_trailers-1;    /* this is the longest one we've got */
         else
            tmostfreq = opt.pref_tlen-1;   /* this is the preferred trailer */
      }
      else {
         if (opt.pref_tlen < tmostfreq+1)
            tmostfreq = opt.pref_tlen-1;   /* only if longer one isn't more frequent */
      }
   }
   else {
      /*
       *  Since we have no particular preference as to the length of the
       *  trailer, check if the chosen trailer starts with a fillchar.
       *  If so, choose a shorter one. Do not eliminate the trailer. :)
       */
      for (cur_tlen = tmostfreq; cur_tlen > 0; --cur_tlen) {
         if (*(trailers[cur_tlen]->text) == opt.fillchar)
            --tmostfreq;
         else
            break;
      }
   }
   /*
    *  Copy chosen trailer, then destruct trailer list
    */
   used_trail->text = (char *) strdup (trailers[tmostfreq]->text);
   used_trail->len  = trailers[tmostfreq]->len;
   used_trail->freq = trailers[tmostfreq]->freq;

   for (cur_tlen=0; cur_tlen<longest; ++cur_tlen) {
      free (trailers[cur_tlen]);
      trailers[cur_tlen] = NULL;
   }
   free (trailers);
   trailers = NULL;

   if (used_trail->freq < 2) do_nothing = true;
}


/****************************************************************************/

void do_stuff_to_lines()
/*
 *  Cut off the trailer from each line. Then remove the now trailing
 *  whitespace. After that, expand tab chars and store expanded version
 *  at text pointer. Deactivate lines that do not meet the maxalign rule
 *  after expansion.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   unsigned long i;
   char *p;
   char *temp, *temp2;
   int pot_newlen;

   for (i=0; i<anz_lines; ++i) {
      /*
       *  If the line is inactive, we will never change it.
       */
      if (lines[i]->active == false) {
         lines[i]->text = lines[i]->orgtext;
         continue;
      }
      /*
       *  If the line does not contain a trailer, we will never change it.
       */
      if (strcmp ((lines[i]->tend)+1-(used_trail->len), used_trail->text)) {
         lines[i]->text = lines[i]->orgtext;
         lines[i]->active = false;
         continue;
      }
      /*
       *  Cut off trailer and then trailing whitespace.
       *  Since we might not really want that, use a temp copy.
       */
      temp = (char *) strdup (lines[i]->orgtext);
      pot_newlen = lines[i]->len;
      p = temp + strlen(temp) - (used_trail->len);
      *p-- = '\0';
      pot_newlen -= used_trail->len;
      while (p >= temp
       && (*p == ' ' || *p == '\t' || *p == '\n')) {
         *p-- = '\0';
         --pot_newlen;
      }
      /*
       *  If the line would still be longer than maxalign, deactivate it.
       */
      if (expand_tabs_into (temp, &temp2) + opt.padding + (used_trail->len)
       > opt.maxalign) {
         lines[i]->text = lines[i]->orgtext;
         lines[i]->active = false;
         free (temp);
         free (temp2);
         continue;
      }
      free (temp2);
      /*
       *  Line length is okay, so bieg everything wieder gerade.
       */
      free (lines[i]->orgtext);
      lines[i]->orgtext = temp;
      lines[i]->len = pot_newlen;
      if (p < temp)
         lines[i]->tend = lines[i]->orgtext;
      else
         lines[i]->tend = p;
      temp = NULL;
      /*
       *  Now expand the tabs and store result at text pointer.
       *  From now on, the len field gives the length of text, not orgtext
       *  for all active lines. len of inactive lines is undefined.
       */
      if (strchr (lines[i]->orgtext, '\t') == NULL)
         lines[i]->text = lines[i]->orgtext;
      else
         lines[i]->len = expand_tabs_into
            (lines[i]->orgtext, &(lines[i]->text));
   }
}


/****************************************************************************/

void output_aligned_lines()
/*
 *  This takes the liste of lines that have been stripped of their trailer
 *  and the then trailing whitespace, and produces the complete aligned
 *  output by filling in fillchars between text and trailer.
 *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */
{
   int newlen;                           /* length of lines to be aligned (w/o trailer) */
   unsigned long i;                      /* loop counter */
   char *sfill;                          /* string with fill characters */
   char *fp;                             /* pointer to zero byte in sfill */
   int fillsize;                         /* number of fillchars in current line */

   /*
    *  If we are not supposed to do anything, just print the lines and go.
    */
   if (do_nothing) {
      for (i=0; i<anz_lines; ++i)
         printf ("%s\n", lines[i]->orgtext);
      return;
   }
   /*
    *  Determine final line length
    */
   if (opt.column > 0                    /* user wants special column */
    && opt.column <= opt.maxalign)
      newlen = opt.column;
   else
      newlen = longest_line_len() + used_trail->len + opt.padding;
   /*
    *  Set up fill string
    */
   sfill = (char *) malloc (newlen+1);
   if (sfill == NULL) {
      perror (PROJECT);
      exit (EXIT_FAILURE);
   }
   memset (sfill, opt.fillchar, newlen);
   fp = sfill;

   /*
    * Output untagged lines, adjust and output tagged ones
    */
   for (i=0; i<anz_lines; ++i) {

      if (lines[i]->active == false) {
         printf ("%s\n", lines[i]->orgtext);
         continue;
      }

      *fp = opt.fillchar;
      fillsize = newlen - lines[i]->len - used_trail->len;
      if (fillsize < 0) fillsize = 0;
      fp = sfill + fillsize;
      *fp = '\0';

      printf ("%s%s%s\n", lines[i]->orgtext, sfill, used_trail->text);
   }
}



/****************************************************************************\
 *  t a l    m a i n ()                                                     *
\****************************************************************************/

int main (int argc, char *argv[]) {

   process_commandline (argc, argv);

   read_all_input();
   find_a_good_trailer();
   if (!do_nothing) do_stuff_to_lines();
   output_aligned_lines();

   return (EXIT_SUCCESS);
}

/*EOF*/
